var __defProp = Object.defineProperty;
var __defProps = Object.defineProperties;
var __getOwnPropDescs = Object.getOwnPropertyDescriptors;
var __getOwnPropSymbols = Object.getOwnPropertySymbols;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __propIsEnum = Object.prototype.propertyIsEnumerable;
var __defNormalProp = (obj, key, value) => key in obj ? __defProp(obj, key, { enumerable: true, configurable: true, writable: true, value }) : obj[key] = value;
var __spreadValues = (a, b) => {
  for (var prop in b || (b = {}))
    if (__hasOwnProp.call(b, prop))
      __defNormalProp(a, prop, b[prop]);
  if (__getOwnPropSymbols)
    for (var prop of __getOwnPropSymbols(b)) {
      if (__propIsEnum.call(b, prop))
        __defNormalProp(a, prop, b[prop]);
    }
  return a;
};
var __spreadProps = (a, b) => __defProps(a, __getOwnPropDescs(b));
var __publicField = (obj, key, value) => {
  __defNormalProp(obj, typeof key !== "symbol" ? key + "" : key, value);
  return value;
};
import { createBrowserHistory } from "history";
export { createBrowserHistory, createHashHistory, createMemoryHistory } from "history";
import { compose, EventEmitter, warningOnce, joinPaths, stripBasename, matchRoutes, isNil } from "oh-router-shared";
class Middleware {
  register(ctx) {
    return true;
  }
}
class RouterMiddleware {
  constructor(middlewares = []) {
    __publicField(this, "middlewares", []);
    this.middlewares = middlewares || [];
  }
  addMiddleware(middlewares) {
    middlewares = Array.isArray(middlewares) ? middlewares : [middlewares];
    this.middlewares.push(...middlewares);
    return this;
  }
  async execMiddlewares(ctx, middlewares) {
    const fn = compose(middlewares.map((m) => {
      return async function(ctx2, next) {
        if (m.register(ctx2)) {
          m.handler(ctx2, next);
        } else {
          next();
        }
      };
    }));
    return fn(ctx);
  }
}
class Router extends RouterMiddleware {
  constructor(opts) {
    super(opts.middlewares);
    __publicField(this, "routes");
    __publicField(this, "history");
    __publicField(this, "basename");
    __publicField(this, "lastHistoryUpdate");
    __publicField(this, "location");
    __publicField(this, "event", new EventEmitter());
    this.routes = opts.routes;
    this.basename = opts.basename || "/";
    this.history = opts.history || createBrowserHistory();
    return this;
  }
  start() {
    if (this.lastHistoryUpdate)
      return this;
    this.history.listen(this.onHistoryLocationChange.bind(this));
    this.onHistoryLocationChange({
      action: this.history.action,
      location: this.history.location
    });
    return this;
  }
  async onHistoryLocationChange(update, force = false) {
    var _a, _b;
    warningOnce("basename", !(this.basename !== "/" && update.location.pathname === "/"), "maybe you should navigate to " + this.basename);
    if (this.lastHistoryUpdate && !force) {
      if (joinPaths([
        this.lastHistoryUpdate.location.pathname,
        this.lastHistoryUpdate.location.search
      ]) === joinPaths([update.location.pathname, update.location.search]))
        return;
    }
    this.lastHistoryUpdate = update;
    let trailingPathname = stripBasename(update.location.pathname, this.basename);
    if (trailingPathname == null) {
      return;
    }
    const matchedRoutes = (_a = matchRoutes(this.routes, {
      pathname: trailingPathname
    })) == null ? void 0 : _a.map((match) => Object.assign({}, match, {
      params: Object.assign({}, match.params)
    }));
    if (isNil(matchedRoutes))
      return;
    if (matchedRoutes[matchedRoutes.length - 1].route.redirect) {
      this.navigate(matchedRoutes[matchedRoutes.length - 1].route.redirect, {
        replace: true
      });
      return;
    }
    const location = __spreadProps(__spreadValues({}, update.location), {
      matched: matchedRoutes,
      meta: matchedRoutes.reduce((pre, cur) => Object.assign(pre, cur.route.meta), {}),
      params: Object.assign({}, (_b = matchedRoutes[matchedRoutes.length - 1]) == null ? void 0 : _b.params),
      action: update.action
    });
    const ctx = {
      to: location,
      from: this.location
    };
    await this.execMiddlewares(ctx, [
      ...this.middlewares,
      {
        handler: async ({ to }) => {
          if (this.lastHistoryUpdate.location.pathname !== to.pathname)
            return;
          this.location = to;
          this.event.emit("onLocationChange", to);
        },
        register: () => true
      }
    ]);
  }
  addLocationListener(listener) {
    this.event.on("onLocationChange", listener);
    if (this.location)
      listener(this.location);
    return this;
  }
  removeLocationListener(listener) {
    this.event.remove("onLocationChange", listener);
  }
  rematch() {
    if (this.lastHistoryUpdate) {
      this.onHistoryLocationChange(this.lastHistoryUpdate, true);
    }
  }
  setRoutes(routes) {
    this.routes = routes;
    return this;
  }
  getRoutes() {
    return this.routes;
  }
  navigate(to, opts) {
    if (typeof to === "number") {
      this.history.go(to);
    } else {
      const _navigate = (opts == null ? void 0 : opts.replace) ? this.history.replace : this.history.push;
      const path = joinPaths([this.basename, to]);
      _navigate(path);
    }
  }
  back() {
    this.navigate(-1);
  }
}
export { Middleware, Router, Router as default };
